/*
 *  SparseImage - Provide a contiguous image file synthesized from
 *  many small chunks for easy backup with rsync etc. using FUSE.
 *  Copyright (C) 2008  Martin Pohlack <martinp (at) gmx.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/* Todo:
   - lock-file against double mounting etc.?
   - useless with encryption on top:
     - support compression of single blocks (gzip etc.)
     - freeing of unused (zeroed) blocks
 */

#define _FILE_OFFSET_BITS 64
#define _XOPEN_SOURCE 500
#define FUSE_USE_VERSION 26

#include <alloca.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <limits.h>
#include <stdint.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <syslog.h>
#include <stdbool.h>
#include <sys/ioctl.h>
#include <linux/fs.h>

#include <fuse.h>

#include "utils.h"

static int chunkdir_fd;
static off_t chunk_size, chunk_count, image_size;

static int ensure_chunkdir_is_cwd()
{
  static bool done;

  if (! done)
  {
      if (fchdir(chunkdir_fd) < 0)
          return -errno;
      done = true;
  }
  return 0;
}

static int resolve_path(const char *path, mode_t *mode)
{
    if (! strcmp(path,"/"))
    {
        *mode = S_IFDIR;
    }
    else if (! strcmp(path, "/image"))
    {
        *mode = S_IFREG;
    }
    else
    {
        return -ENOENT;
    }
    return 0;
}

static void gen_chunk_name(char *buf, off_t num)
{
    sprintf(buf, "%llx", num);  // flat chunk names, in hex
}

static int sparse_image_getattr(const char *path, struct stat *buf)
{
    int r;
    mode_t mode;

    if ((r = resolve_path(path, &mode)) < 0)
        return r;
    if (fstat(chunkdir_fd, buf) < 0)
        return -errno;
    buf->st_mode = (buf->st_mode & ~ (S_IFMT | S_IXUSR | S_IXGRP | S_IXOTH)) |
                                 (mode & S_IFMT);
    if (S_ISDIR(mode))
    {
        buf->st_mode = buf->st_mode |((buf->st_mode & S_IRUSR)?S_IXUSR : 0)
                                    |((buf->st_mode & S_IRGRP)?S_IXGRP : 0)
                                    |((buf->st_mode & S_IROTH)?S_IXOTH : 0);
        buf->st_nlink = 2;
        buf->st_size = 0;
    }
    else
    {
        buf->st_nlink = 1;
        buf->st_size = image_size;
    }
    buf->st_blocks = 0;
    return 0;
}

static int sparse_image_readdir(const char *path, void *buf,
                                fuse_fill_dir_t filler, off_t offset,
                                struct fuse_file_info *fi)
{
    int r;
    mode_t mode;

    // printf("readdir()\n");

    if ((r = resolve_path(path, &mode)) < 0)
        return r;
    if (! S_ISDIR(mode))
        return -ENOTDIR;
    filler(buf, ".", NULL, 0);
    filler(buf, "..", NULL, 0);
    filler(buf, "image", NULL, 0);
    return 0;
}

static int sparse_image_open(const char *path, struct fuse_file_info *fi)
{
    int r;
    mode_t mode;

    // printf("open()\n");

    if ((r = resolve_path(path, &mode)) < 0)
        return r;
    return 0;
}

ssize_t read_from_chunk(off_t chunk_num, char * buf, size_t count,
                        off_t offset)
{
    int chunk_fd;
    char chunk_name[24];
    ssize_t r, c = count;

    gen_chunk_name(chunk_name, chunk_num);
    chunk_fd = open(chunk_name, O_RDONLY);
    if (chunk_fd < 0)
    {
        if (errno == ENOENT)  // no file
        {
            memset(buf, 0, count);  // zero all
            return count;
        }
        else
        {
            return -EIO;
        }
    }

    while (c)
    {
        r = pread(chunk_fd, buf, c, offset);
        if (-1 == r && errno == EINTR)
            continue;
        if (r < 0)
        {
            if (close(chunk_fd) < 0)
                die(true, "close(chunk_fd)");

            return -errno;
        }
        if (0 == r)
            break;  // EOF
        c      -= r;
        offset += r;
        buf    += r;
    }

    memset(buf, 0, c);  // zero rest

    if (close(chunk_fd) < 0)
        die(true, "close(chunk_fd)");

    return count;
}

static int sparse_image_read(const char *path, char *buf, size_t count,
                             off_t offset, struct fuse_file_info *fi)
{
    int r;
    mode_t mode;
    off_t chunk_num, chunk_offset;
    ssize_t got = 0, r2;
    size_t current_count;

    // printf("read(%s, %d, %lld)\n", path, count, offset);

    if (ensure_chunkdir_is_cwd() < 0)
        return -EIO;
    if ((r = resolve_path(path, &mode)) < 0)
        return r;
    if (S_ISDIR(mode))
        return -EISDIR;
    if (count > INT_MAX)
        die(false, "read request larger than can be represented in an int");
    if (count > SSIZE_MAX)
        die(false, "read request larger than can be represented in an ssize_t");
    count = min((off_t)count, max(image_size - offset, (off_t)0));

    while (count > 0)
    {
        chunk_num     = offset / chunk_size;
        if (chunk_num > chunk_count)
            break;  // EOF
        chunk_offset  = offset % chunk_size;
        current_count = min(count, ((chunk_num + 1) * chunk_size) - offset);
        r2 = read_from_chunk(chunk_num, buf + got, current_count, chunk_offset);
        if (r2 < 0)
            return r2;
        if (0 == r2)  // end of device
            break;
        count -= r2;
        offset += r2;
        got += r2;
    }
    return got;
}

ssize_t write_to_chunk(off_t chunk_num, const char * buf, size_t count,
                       off_t offset)
{
    int chunk_fd;
    char chunk_name[24];
    ssize_t w;
    size_t c = count;

    gen_chunk_name(chunk_name, chunk_num);

    chunk_fd = open(chunk_name, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
    if (chunk_fd < 0)
    {
        return -EIO;
    }

    while (c)
    {
        w = pwrite(chunk_fd, buf, c, offset);
        if (-1 == w && errno == EINTR)
            continue;
        if (w < 0)
        {
            if (close(chunk_fd) < 0)
                die(true, "close(chunk_fd)");

            return -errno;
        }
        c      -= w;
        offset += w;
        buf    += w;

    }

    if (close(chunk_fd) < 0)
        die(true, "close(chunk_fd)");

    return count;
}

static int sparse_image_write(const char *path, const char *buf, size_t count,
                              off_t offset, struct fuse_file_info *fi)
{
    int r;
    mode_t mode;
    off_t chunk_num, chunk_offset;
    size_t got = 0, w;
    size_t current_count;

    if (ensure_chunkdir_is_cwd() < 0)
        return -EIO;
    if ((r = resolve_path(path, &mode)) < 0)
        return r;
    if (S_ISDIR(mode))
        return -EISDIR;
    if (count > INT_MAX)
        die(false, "read request larger than can be represented in an int");
    if (count > SSIZE_MAX)
        die(false, "read request larger than can be represented in an ssize_t");
    count = min((off_t)count, max(image_size - offset, (off_t)0));

    while (count > 0)
    {
        chunk_num     = offset / chunk_size;
        if (chunk_num > chunk_count)
            break;  // EOF
        chunk_offset  = offset % chunk_size;
        current_count = min(count, ((chunk_num + 1) * chunk_size) - offset);
        w = write_to_chunk(chunk_num, buf + got, current_count, chunk_offset);
        if (w < 0)
            return w;
        if (0 == w)  // EOF
            break;
        count -= w;
        offset += w;
        got += w;
    }
    return got;
}

static struct fuse_operations sparse_image_ops = {
    .getattr = sparse_image_getattr,
    .readdir = sparse_image_readdir,
    .open    = sparse_image_open,
    .read    = sparse_image_read,
    .write   = sparse_image_write
};

static void parse_config_file(char * dir, off_t *chunk_size,
                              off_t *chunk_count)
{
    FILE * file;
    char * prop_file_name;
    char line[1024];
    char key[101];
    off_t val;
    int ret;

    prop_file_name = alloca(strlen(dir) + 16);
    strcpy(prop_file_name, dir);
    strcat(prop_file_name, "/metadata.prop");

    file = fopen(prop_file_name, "r");
    if (NULL == file)
        die(true, "open(<chunk dir>/metadata.prop)");

    while (fgets(line, 1024, file))
    {
        ret = sscanf(line, "%100[^#=]=%lld", key, &val);
        if (ret != 2)
            continue;

        if (0 == strcmp("chunk_size", key))
            *chunk_size = val;
        else if (0 == strcmp("chunk_count", key))
            *chunk_count = val;
        else
            printf("Unknown property '%s', ignored!\n", key);
    }

    fclose(file);
}

int main(int argc, char **argv)
{
    char opt;
    char *helpargv[2] = {NULL, "-h"};
    char *chunks_dir_name;

    openlog("sparse_image", LOG_CONS | LOG_NDELAY | LOG_PERROR | LOG_PID,
            LOG_DAEMON);
    opterr = 0;
    while ((opt = getopt(argc, argv, "ho:")) != -1)
        if (opt == 'h')
        {
            helpargv[0] = argv[0];
            return fuse_main(2, helpargv, NULL, NULL);
        }
    if (argc - optind != 2)
        die(false, "Usage: %s [options] <chunk dir> <mount point>", *argv);

    chunks_dir_name = alloca(strlen(argv[optind]) + 10);
    strcpy(chunks_dir_name, argv[optind]);
    strcat(chunks_dir_name, "/chunks/");
    if ((chunkdir_fd = open(chunks_dir_name, O_RDONLY)) < 0)
        die(true, "open(<chunk dir>/chunks/)");

    // assumed some default values
    chunk_size = 1024 * 1024;
    chunk_count = 10;
    // no check for the real stuff
    parse_config_file(argv[optind], &chunk_size, &chunk_count);

    image_size = chunk_size * chunk_count;

    argv[optind] = argv[optind + 1];
    argc--;
    return fuse_main(argc, argv, &sparse_image_ops, NULL);
}
